*      -----------------------------------------------------------
*     | The Power Series Solutions of the Campbell-Cochrane Model |
*     |         Yu Chen, Tom Cosimano, and Alex Himonas           |
*     |                       July 1, 2007                        |
*      -----------------------------------------------------------

*     REMARK: All the formulas used in this program are derived in
*     "Numerical Scheme for Solving the CC-Model".

*     --- Main Program ---

      PROGRAM MAIN

*      -----------------------------------------------------------
*     |  Define the radius r of interval around zero for Q(x).    |
*      -----------------------------------------------------------

      PARAMETER (r=0.25)

*      -----------------------------------------------------------
*     |  Define the order of the Taylor polynomial approximation  |
*     |  of the function Q(x) around zero, denoted by N_coeff > 2.|
*      -----------------------------------------------------------

      PARAMETER (N_coeff=50)

*      -----------------------------------------------------------
*     |  Define the number of terms to be used in the partial sum |
*     |  approximations to the series, denoted by N_term > 2.     |
*      -----------------------------------------------------------

      PARAMETER (N_term=80)

*      -----------------------------------------------------------
*     |  Compute the ranges of the indices in b[s,t,l], where     |
*     |     0 <= s <= N1, 0 <= t <= N2, and 0 <= l <= N3.         |
*      -----------------------------------------------------------

      PARAMETER (N1=2*N_term+N_coeff+1)
      PARAMETER (N2=2*N_term+1)
      PARAMETER (N3=N_coeff)

*      -----------------------------------------------------------
*     |  Compute the number of binomial coefficients to be used.  |
*     |                   Number=N_bi_coeff                       |
*      -----------------------------------------------------------

      PARAMETER (N_bi_coeff=(N1+1)*(N1+2)/2)

*      -----------------------------------------------------------
*     |  Declare the types of variables to be used in the program.|
*     |  All the floating-point compuations use DOUBLE PRECISION, |
*     |  i.e., sixteen-digit floating-point representations.      |
*      -----------------------------------------------------------

*     *** List of INTEGER Variables ***

      INTEGER  i, j, k, l, n, s, t 
      INTEGER  Index, Index1, Index2

*     *** List of DOUBLE PRECISION variables ***

      DOUBLE PRECISION  bet, c, G, gam, phi, S_bar, sig
      DOUBLE PRECISION  K0, K1, K2, K3
      DOUBLE PRECISION  A_temp, fac, Gamk, Neg, PI, temp, TWO

*     *** List of DOUBLE PRECISION arrays ***
*     (1)  Bi_coeff[1..N_bi_coeff]=the binomial coefficients.
*     (2)  lam[0..N_coeff]=the derivatives of the lamdba function.
*     (3)  Lamk[0..2*N_term+1,0..N_coeff]=the derivatives of the powers
*          of the lambda function.
*     (4)  Inv_Lamk[0..2*N_term+1,0..N_coeff]=the derivatives of the
*          inverses of the powers of the lambda function.
*     (5)  psi[0..N_coeff]=the derivatives of the psi function.
*     (6)  Psik[0..2*N_term+N_ceoff+1,0..N_coeff]=the derivatives of
*          the powers of the psi function.
*     (7)  b[0..N1,0..N2,0..N3]=the derivatives of the powers of the
*          psi function over the powers of the lambda function.
*     (8)  A[0..N_coeff,0..N_coeff]=the coefficient matrix of the 
*          linear system.
*     (9)  f[0..N_coeff]=the constants of the linear system.
*     (10) u[0..N_coeff]=the solution of the linear system.
*     (11) rk[0..2*N_term+N_coeff+1]=the powers of r.
*     *** ****************************** ***

      DOUBLE PRECISION  Bi_coeff(N_bi_coeff)
      DOUBLE PRECISION  lam(N3+1), Lamk(N2+1,N3+1), Inv_Lamk(N2+1,N3+1)
      DOUBLE PRECISION  psi(N3+1), Psik(N1+1,N3+1)
      DOUBLE PRECISION  b(N1+1,N2+1,N3+1) 
      DOUBLE PRECISION  A(N3+1,N3+1), f(N3+1), u(N3+1)
      DOUBLE PRECISION  rk(N1+1)

*     *** List of CHARACTER variables ***

      CHARACTER  Ans

*     *** The Name of Data File to Be Created ***

      CHARACTER*12  Filename

*      ----------------------------------------------------------
*     |  Draw a straight line.                                   |
*      ---------------------------------------------------------

      CALL CLRSCN(2)

      WRITE(*,1000)
 1000 FORMAT(1X, 15X, 'The Numerical Solutions of the CC-Model')


      WRITE(*,1010)
 1010 FORMAT(1X, 70('-'))

*      -----------------------------------------------------------
*     |  Define the basic parameters: bet, G, gam, S_bar, sig.    |
*      -----------------------------------------------------------

      bet=0.99
      G=1.001575
      gam=2.0
      S_bar=0.057
      sig=0.015/SQRT(12.0)

*      -----------------------------------------------------------
*     |  Calculate the derived parameter: phi, c, K0, K1, K2, K3. |
*      -----------------------------------------------------------

      phi=1-(gam*(sig**2))/(S_bar**2)
      c=SQRT((1-phi)/gam)
      K0=bet*(G**(1-gam))
      K1=(gam*sig/S_bar)**2
      K2=(0.112*0.2)/(sig*SQRT(12.0))
      K3=K0*EXP(((sig*(K2-gam))**2)/2)

*      -----------------------------------------------------------
*     |  Define the value of pi, denoted by PI.                   |
*      -----------------------------------------------------------

      PI=3.141592653589793238462643

*      -----------------------------------------------------------
*     |  Step 1: Calculate the binomial coefficients.             |
*      -----------------------------------------------------------

      WRITE(*,*)

      WRITE(*,1020) 
 1020 FORMAT(2X, '1. The binomial coefficients are calculated.')

      Bi_coeff(1)=1
      Bi_coeff(2)=1
      Bi_coeff(3)=1
      DO 10 i=2,N1
         Bi_coeff(i*(i+1)/2+1)=1
         DO 20 j=1,(i-1)
            Bi_coeff((i*(i+1)/2+1)+j)=Bi_coeff(((i-1)*i/2+1)+(j-1))
     C                               +Bi_coeff(((i-1)*i/2+1)+j)
 20      CONTINUE
         Bi_coeff((i*(i+1)/2+1)+i)=1
 10   CONTINUE

*      -----------------------------------------------------------
*     |  Step 2: Calculate the derivatives of the lambda function,|
*     |  where the lambda function is the sensitivity to the      |
*     |  surplus consumption groweh in the CC-model.              |
*      -----------------------------------------------------------

      WRITE(*,1030)
 1030 FORMAT(2X, 
     C '2. The derivatives of the lambda function are calculated.')
      lam(0+1)=c-sig
      lam(1+1)=-c
      DO 30 i=2,N_coeff
         lam(i+1)=(2*i-3)*lam((i-1)+1)
 30   CONTINUE

*      -----------------------------------------------------------
*     |  Step 3: Calculate the derivatives of the powers of the   |
*     |  lambda function.                                         |
*      -----------------------------------------------------------

      WRITE(*,1040)
 1040 FORMAT(2X,
     C '3. The derivatives of the powers of lambda are calculated.')
      Lamk(0+1,0+1)=1
      DO 40 j=1,N_coeff
         Lamk(0+1,j+1)=0
 40   CONTINUE
      DO 50 j=0,N_coeff
         Lamk(1+1,j+1)=lam(j+1)
 50   CONTINUE
      DO 60 i=2,(2*N_term+1)
         Lamk(i+1,0+1)=lam(0+1)**i
         DO 70 j=1,N_coeff
            Lamk(i+1,j+1)=0
            DO 80 k=0,j
               Lamk(i+1,j+1)=Lamk(i+1,j+1)+Bi_coeff((j*(j+1)/2+1)+k)
     C         *Lamk((i-1)+1,k+1)*lam((j-k)+1)
 80         CONTINUE
 70      CONTINUE
 60   CONTINUE
         
*      ----------------------------------------------------------
*     |  Step 4: Calculate the derivatives of the inverses of the|
*     |  powers of the lambda function.                          |
*      ----------------------------------------------------------

      WRITE(*,1050)
 1050 FORMAT(2X, '4. The derivatives of the inverses of the powers of',
     C ' lambda are calculated.')

      Inv_Lamk(0+1,0+1)=1
      DO 90 j=1,N_coeff
         Inv_Lamk(0+1,j+1)=0
 90   CONTINUE
      DO 100 k=0,(2*N_term+1)
         Inv_Lamk(k+1,0+1)=1/Lamk(k+1,0+1)
         DO 110 j=1,N_coeff
            Inv_Lamk(k+1,j+1)=0
            DO 120 i=0,(j-1)
               Inv_Lamk(k+1,j+1)=Inv_Lamk(k+1,j+1)
     C         +Bi_coeff((j*(j+1)/2+1)+i)*Inv_Lamk(k+1,i+1)
     C         *Lamk(k+1,(j-i)+1)
 120        CONTINUE
            Inv_Lamk(k+1,j+1)=-Inv_Lamk(k+1,j+1)/Lamk(k+1,0+1)
 110     CONTINUE
 100  CONTINUE

*      ----------------------------------------------------------
*     |  Step 5: Calculate the derivatives of the psi function.  |
*      ----------------------------------------------------------

      WRITE(*,1060)
 1060 FORMAT(2X, '5. The derivatives of the psi function are',
     C  ' calculated.')

      psi(0+1)=sig*(K2-gam)*(c-sig)
      psi(1+1)=phi-sig*(K2-gam)*c
      psi(2+1)=-sig*(K2-gam)*c
      DO 130 i=3,N_coeff
         psi(i+1)=(2*i-3)*psi((i-1)+1)
 130  CONTINUE
      
*      -----------------------------------------------------------
*     |  Step 6: Calculate the derivatives of the powers of the   |
*     |  psi function.                                            |
*      -----------------------------------------------------------

      WRITE(*,1070)
 1070 FORMAT(2X, '6. The derivatives of the powers of the psi',
     C ' function are calculated.')

      Psik(0+1,0+1)=1
      DO 140 j=1,N_coeff
         Psik(0+1,j+1)=0
 140  CONTINUE
      DO 150 j=0,N_coeff
         Psik(1+1,j+1)=psi(j+1)
 150  CONTINUE
      DO 160 k=2,(2*N_term+N_coeff+1)
         Psik(k+1,0+1)=psi(0+1)**k
         DO 170 j=1,N_coeff
            Psik(k+1,j+1)=0
            DO 180 i=0,j
               Psik(k+1,j+1)=Psik(k+1,j+1)+Bi_coeff((j*(j+1)/2+1)+i)
     C         *Psik((k-1)+1,i+1)*psi((j-i)+1)
 180        CONTINUE
 170     CONTINUE
 160  CONTINUE

*      -----------------------------------------------------------
*     |  Step 7: Calculate the array b[s,t,l]                     |
*      -----------------------------------------------------------

      WRITE(*,1080) 
 1080 FORMAT(2X, '7. The array b[s,t,l] is calculated.')

      DO 190 s=0,N1
         DO 200 t=0,N2
            DO 210 l=0,N3
               b(s+1,t+1,l+1)=0
               DO 220 i=0,l
                  b(s+1,t+1,l+1)=b(s+1,t+1,l+1)
     C            +Bi_coeff((l*(l+1)/2+1)+i)
     C            *Psik(s+1,i+1)*Inv_Lamk(t+1,(l-i)+1)
                  IF (i.EQ.0) THEN
                     fac=1
                  ELSE
                     fac=fac*i
                  ENDIF
 220           CONTINUE
               b(s+1,t+1,l+1)=b(s+1,t+1,l+1)/fac
 210        CONTINUE
 200     CONTINUE
 190  CONTINUE

*      -----------------------------------------------------------
*     |  Step 8: Calculate the powers of r.                       |
*      -----------------------------------------------------------

      WRITE(*,1090)
 1090 FORMAT(2X, '8. The powers of r are calculated.')

      rk(0+1)=1
      DO 230 i=1,(2*N_term+N_coeff+1)
         rk(i+1)=rk((i-1)+1)*r
 230  CONTINUE

*      -----------------------------------------------------------
*     |  Step 9: Calculate the coefficient matrix of the system.  |
*      -----------------------------------------------------------

      WRITE(*,1100) 
 1100 FORMAT(2X, '9. The coefficient matrix is constructed.')

      DO 240 l=0,N_coeff
         DO 250 n=0,N_coeff
            A(l+1,n+1)=0
            DO 260 k=0,N_term
               IF (k.EQ.0) THEN
                  TWO=1
                  fac=1
               ELSE
                  TWO=TWO*2
                  fac=fac*k
               ENDIF
               A_temp=0
               DO 270 j=0,n
                  DO 280 i=0,(2*k+j+1)
                     CALL PARITY(j+k,Index1)
                     CALL PARITY(i+j+k+1,Index2)
                     IF (Index1.EQ.Index2) THEN
                        temp=Bi_coeff((n*(n+1)/2+1)+j)
                        temp=temp*Bi_coeff(((2*k+j+1)
     C                       *((2*k+j+1)+1)/2+1)+i)
                        temp=temp*((-1)**(Index1)+(-1)**(Index2))
     C                       *rk(i+1)/(2*k+j+1)
                        A_temp=A_temp+temp
     C                       *b((2*k+1+n-i)+1,(2*k+1)+1,l+1)
                     ENDIF
 280              CONTINUE
 270           CONTINUE
               DO 290 j=0,(2*k+1)
                  CALL PARITY(j+k,Index1)
                  CALL PARITY(n+k+1,Index2)
                  IF (Index1.EQ.Index2) THEN
                     temp=Bi_coeff(((2*k+1)*((2*k+1)+1)/2+1)+j)
                     temp=temp*((-1)**(Index1)+(-1)**(Index2))
     C                    *rk((n+j)+1)/(2*k+1)
                     A_temp=A_temp+temp*b((2*k+1-j)+1,(2*k+1)+1,l+1)
                  ENDIF
 290           CONTINUE
               A_temp=A_temp*K3/(SQRT(2*PI)*TWO*fac)
               A(l+1,n+1)=A(l+1,n+1)+A_temp
 260        CONTINUE
            CALL PARITY(n,Index)
            IF ((l.EQ.0).AND.(Index.EQ.0)) THEN
               A(l+1,n+1)=A(l+1,n+1)+K3*rk(n+1)
            ENDIF
            IF (l.EQ.n) THEN
               A(l+1,n+1)=A(l+1,n+1)-1
            ENDIF
 250     CONTINUE
 240  CONTINUE
             
*      ----------------------------------------------------------
*     |  Step 10: Calculate the constants in the system.         |
*      ----------------------------------------------------------

      WRITE(*,1110) 
 1110 FORMAT(1X, '10. The constants of the system are calculated.')

      DO 300 l=0,N_coeff
         IF (l.EQ.0) THEN
            fac=1
            Gamk=1
            Neg=1
         ELSE
            fac=fac*l
            Gamk=Gamk*gam
            Neg=-Neg
         ENDIF
         f(l+1)=-Neg*Gamk/fac
 300  CONTINUE

*      -----------------------------------------------------------
*     |  Step 11: Solve the linear system.                        |
*      -----------------------------------------------------------
 
      WRITE(*,1120) 
 1120 FORMAT(1X, '11. The linear system is solved.')

      CALL LinearSolver(N_coeff+1, A, f, u)

*      -----------------------------------------------------------
*     |  Print all the parameters used in the program.            |
*      -----------------------------------------------------------

      CALL CLRSCN(3)

      WRITE(*,2010)
 2010 FORMAT(1X, 70('-'))
      WRITE(*,2020) 
 2020 FORMAT(1X, 30X, 'Results')

      WRITE(*,*)

      WRITE(*,2030) 
 2030 FORMAT(1X, 'Basic Parameters:')
      WRITE(*,2040) bet, G, gam, S_bar, sig
 2040 FORMAT(1X, 3X, 'bet=', F8.5, 3X,'G=', F8.5, 3X, 
     C       'gam=', F8.5, 3X, 'S_bar=', F8.5, 3X, 'sig=',F8.5)

      WRITE(*,*)

      WRITE(*,2050)
 2050 FORMAT(1X, 'Derived Parameters:')
      WRITE(*,2060) c, phi
 2060 FORMAT(1X, 3X, 'c=', F8.5, 5X, 'phi=', F8.5)
      WRITE(*,2070) K0, K1, K2, K3
 2070 FORMAT(1X, 3X, 'K0=', F8.5, 4X, 'K1=', F8.5, 4X, 'K2=', F8.5,
     C       4X, 'K3=', F8.5)

      WRITE(*,*)

*      -----------------------------------------------------------
*     |  Print the coefficients of the Taylor polynomial.         |
*      -----------------------------------------------------------

      WRITE(*,2080)
 2080 FORMAT(1X, 70('-'))

      WRITE(*,2090) N_coeff
 2090 FORMAT(1X, 8X, 'The Coefficients of', I4,'-th Order Taylor',
     C  ' Polynomial in') 
      WRITE(*,2100) -r, r
 2100 FORMAT(1X, 25X, F6.4, '< x <', F6.4)

      WRITE(*,*)

      WRITE(*,2110)
 2110 FORMAT(1X, 17X, '   Order        Coefficient    ')
      WRITE(*,2120)
 2120 FORMAT(1X, 17X, '   ============================')

      DO 510 i=0,N_coeff
         WRITE(*,2130) i, u(i+1)
 2130    FORMAT(1X, 17X, 2X, I4, 8X, E16.10)
 510  CONTINUE
      
      WRITE(*,2120)

      WRITE(*,*)

*      ------------------------------------------------------------
*     |  Save or do not save the results.                          |
*      ------------------------------------------------------------

 3000 WRITE(*,*) 'Do you want to save the above coefficients?',
     C           ' --- yes (y) or no (n)' 
      READ(*,*) Ans
      IF ((Ans.NE.'y').AND.(Ans.NE.'n')) THEN
         WRITE(*,*) 'It is neither y nor n. Try again!'
         GOTO 3000
      ENDIF
      
      IF (Ans.EQ.'y') THEN
         WRITE(*,*) 'Please enter the name of the data file:'
         READ(*,2200) Filename
 2200    FORMAT(A12)
         OPEN(1, file=Filename, status='NEW', form='FORMATTED',
     C        access='DIRECT', recl=80)
         DO 400 i=0,N_coeff
            WRITE(1,2210,rec=i+1) i, u(i+1)
 2210       FORMAT(1X, 'u(', I4, ')=', E16.10)
 400     CONTINUE
         CLOSE(1)
         WRITE(*,2220) Filename
 2220    FORMAT('The coefficients have been saved in the file ',
     C          '"',A12,'".')
      ENDIF

      WRITE(*,*)
      WRITE(*,2230)
 2230 FORMAT(1X,30X,'Finished!')
      CALL CLRSCN(2)

      END


*     --- SUBROUTINES ---

*      -----------------------------------------------------------
*     |  SUBROUTINE 1: Determine an even or odd integer, where    |
*     |               even interger=0 and odd integer=1.          |
*      -----------------------------------------------------------

      SUBROUTINE  PARITY(N,par)
      INTEGER  N, par

      par=1
      IF ((INT(N/2)*2).EQ.N) THEN 
         par=0
      ENDIF
      RETURN
      END


*      ------------------------------------------------------------
*     |  SUBROUTINE 2: Clear the screen.                           |
*      -----------------------------------------------------------

      SUBROUTINE CLRSCN(N)
      INTEGER  N

      DO 10 I=1,N
 10   WRITE(*,*)   
      RETURN
      END

*      ------------------------------------------------------------
*     |  SUBROUTINE 3: Gaussian Elimination Method                 |
*      ------------------------------------------------------------

      SUBROUTINE LinearSolver(n, AA, ff, X)
      DOUBLE PRECISION  A(n,n+1), AA(n,n), ff(n), X(n)
      DOUBLE PRECISION  c, M, P, V_temp
      INTEGER  I, J, K, I_temp

      DO 100 I=1,n
         DO 110 J=1,n
            A(I,J)=AA(I,J)
 110     CONTINUE
         A(I,n+1)=ff(I)
 100  CONTINUE
   
      DO 10 I=1,(n-1)
         I_temp=I
         M=A(I,I)
         Do 20 K=I+1,n
            IF (ABS(A(K,I)).GT.ABS(M)) THEN
               I_temp=K
               M=A(K,I)
            ENDIF
 20      CONTINUE
         IF (M.EQ.0) THEN
            STOP 'No unique solution exists.'
         ENDIF
         IF ((I_temp).NE.I) THEN
            DO 30 J=I,(n+1)
               V_temp=A(I,J)
               A(I,J)=A(I_temp,J)
               A(I_temp,J)=V_temp
 30         CONTINUE
         ENDIF
         DO 40 K=(I+1),n
            c=A(K,I)/A(I,I)
            A(K,I)=0
            DO 50 J=(I+1),(n+1)
               A(K,J)=A(K,J)-c*A(I,J)
 50         CONTINUE
 40      CONTINUE
 10    CONTINUE
       
       IF (A(n,n).EQ.0) THEN
          STOP 'No unique solution exists.'
       ENDIF

       X(n)=A(n,n+1)/A(n,n)

       DO 60 I=(n-1),1,-1
          X(I)=A(I,n+1)
          DO 70 J=(I+1),n
             X(I)=X(I)-A(I,J)*X(J)
 70       CONTINUE
          X(I)=X(I)/A(I,I)
 60    CONTINUE

       RETURN
       END
